/*
 * libexplain - Explain errno values returned by libc functions
 * Copyright (C) 2008, 2009 Peter Miller
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef LIBEXPLAIN_READDIR_H
#define LIBEXPLAIN_READDIR_H

#include <libexplain/warn_unused_result.h>
#include <libexplain/large_file_support.h>

#include <dirent.h>

/**
  * @file
  * @brief explain readdir(2) errors
  */

#ifdef __cplusplus
extern "C" {
#endif

/**
  * The explain_readdir_or_die function is
  * used to call the readdir(2) system call.  On
  * failure an explanation will be printed to stderr,
  * obtained from explain_readdir(3), and
  * then the process terminates by calling exit(EXIT_FAILURE).
  *
  * This function is intended to be used in a fashion
  * similar to the following example:
  * @code
  * explain_readdir_or_die(dir);
  * @endcode
  *
  * @param dir
  *     The dir, exactly as to be passed to the readdir(2) system call.
  * @returns
  *     a pointer to a dirent structure, or NULL if end-of-file is
  *     reached.  On failure, prints an explanation and exits, it does
  *     not return.
  */
struct dirent *explain_readdir_or_die(DIR *dir);

/**
  * The explain_readdir function is used to
  * obtain an explanation of an error returned by the
  * readdir(2) system call.
  * The least the message will contain is the value of
  * strerror(errno), but usually it will do much better,
  * and indicate the underlying cause in more detail.
  *
  * The errno global variable will be used to obtain the
  * error value to be decoded.
  *
  * This function is intended to be used in a fashion
  * similar to the following example:
  * @code
  * errno = 0;
  * struct dirent *dep = readdir(dir);
  * if (!dep && errno != 0)
  * {
  *     fprintf(stderr, "%s\n", explain_readdir(dir));
  *     exit(EXIT_FAILURE);
  * }
  * @endcode
  *
  * @param dir
  *     The original dir, exactly as passed to the readdir(2) system call.
  * @returns
  *     The message explaining the error.  This
  *     message buffer is shared by all libexplain
  *     functions which do not supply a buffer in their
  *     argument list.  This will be overwritten by the
  *     next call to any libexplain function which shares
  *     this buffer, including other threads.
  * @note
  *     This function is <b>not</b> thread safe, because
  *     it shares a return buffer across all threads, and
  *     many other functions in this library.
  */
const char *explain_readdir(DIR *dir)
                                                  LIBEXPLAIN_WARN_UNUSED_RESULT;

/**
  * The explain_errno_readdir function is
  * used to obtain an explanation of an error returned by
  * the readdir(2) system call.
  * The least the message will contain is the value of
  * strerror(errnum), but usually it will do much better,
  * and indicate the underlying cause in more detail.
  *
  * This function is intended to be used in a fashion
  * similar to the following example:
  * @code
  * errno = 0;
  * struct dirent *dep = readdir(dir);
  * int err = errno;
  * if (!dep && err != 0)
  * {
  *     fprintf(stderr, "%s\n", explain_readdir(err, dir));
  *     exit(EXIT_FAILURE);
  * }
  * @endcode
  *
  * @param errnum
  *     The error value to be decoded, usually obtained
  *     from the errno global variable just before this
  *     function is called.  This is necessary if you need
  *     to call <b>any</b> code between the system call to
  *     be explained and this function, because many libc
  *     functions will alter the value of errno.
  * @param dir
  *     The original dir, exactly as passed to the readdir(2) system call.
  * @returns
  *     The message explaining the error.  This
  *     message buffer is shared by all libexplain
  *     functions which do not supply a buffer in their
  *     argument list.  This will be overwritten by the
  *     next call to any libexplain function which shares
  *     this buffer, including other threads.
  * @note
  *     This function is <b>not</b> thread safe, because
  *     it shares a return buffer across all threads, and
  *     many other functions in this library.
  */
const char *explain_errno_readdir(int errnum, DIR *dir)
                                                  LIBEXPLAIN_WARN_UNUSED_RESULT;

/**
  * The explain_message_readdir function is
  * used to obtain an explanation of an error returned by
  * the readdir(2) system call.
  * The least the message will contain is the value of
  * strerror(errno), but usually it will do much better,
  * and indicate the underlying cause in more detail.
  *
  * The errno global variable will be used to obtain the
  * error value to be decoded.
  *
  * This function is intended to be used in a fashion
  * similar to the following example:
  * @code
  * errno = 0;
  * struct dirent *dep = readdir(dir);
  * if (!dep && errno != 0)
  * {
  *     char message[3000];
  *     explain_message_readdir(message, sizeof(message), dir);
  *     fprintf(stderr, "%s\n", message);
  *     exit(EXIT_FAILURE);
  * }
  * @endcode
  *
  * @param message
  *     The location in which to store the returned
  *     message.  If a suitable message return buffer is
  *     supplied, this function is thread safe.
  * @param message_size
  *     The size in bytes of the location in which to
  *     store the returned message.
  * @param dir
  *     The original dir, exactly as passed to the readdir(2) system call.
  */
void explain_message_readdir(char *message, int message_size, DIR *dir);

/**
  * The explain_message_errno_readdir
  * function is used to obtain an explanation of an error
  * returned by the
  * readdir(2) system call.
  * The least the message will contain is the value of
  * strerror(errnum), but usually it will do much better,
  * and indicate the underlying cause in more detail.
  *
  * This function is intended to be used in a fashion
  * similar to the following example:
  * @code
  * errno = 0;
  * struct dirent *dep = readdir(dir);
  * int err = errno;
  * if (readdir(dir) < 0)
  * {
  *     char message[3000];
  *     explain_message_errno_readdir(message, sizeof(message), err, dir);
  *     fprintf(stderr, "%s\n", message);
  *     exit(EXIT_FAILURE);
  * }
  * @endcode
  *
  * @param message
  *     The location in which to store the returned
  *     message.  If a suitable message return buffer is
  *     supplied, this function is thread safe.
  * @param message_size
  *     The size in bytes of the location in which to
  *     store the returned message.
  * @param errnum
  *     The error value to be decoded, usually obtained
  *     from the errno global variable just before this
  *     function is called.  This is necessary if you need
  *     to call <b>any</b> code between the system call to
  *     be explained and this function, because many libc
  *     functions will alter the value of errno.
  * @param dir
  *     The original dir, exactly as passed to the readdir(2) system call.
  */
void explain_message_errno_readdir(char *message, int message_size,
    int errnum, DIR *dir);

#ifdef __cplusplus
}
#endif

#endif /* LIBEXPLAIN_READDIR_H */
